static void filter_register();

#include "vpi_user.h"
#include <assert.h>
#include <stdio.h>
#include <math.h>

/* these three are state variables for ODE integration */
static double vp=1.0;   /* output of phase detector */
/* near stable values vi=-2.9e-5, vo=0.9 */
static double vi=0;   /* integral of vp           */
static double vo=0;   /* vout (controls VCO)      */
static int last_time=0;

/* returns frequency in MHz given Vtune in Volts */
static double emulate_max2605( double vt )
{
	static double pp[] = { -31.568, 302.434, 184.118 };
	static double rr = 1.19;
	double poly;

	/* In octave, this is written
	 *    (sqrt(polyval(pp,vt))+35)*rr
	 * where rr=1.19 changes scale from curve on data sheet
	 * to our 68 MHz nominal center frequency.
	 */
	poly = (pp[0]*vt+pp[1])*vt+pp[2];
	return (sqrt(poly)+35.0)*rr;
}

static void update_state(struct t_vpi_time *time)
{
	const double A0 = 4.e4;     /*  /second  */
	const double t1 = 14.0e-6;  /*  second   */
	const double t2 = 0.14e-6;  /*  second   */

	double dt = 1.e-12*(time->low - last_time);  /* convert to seconds */
	vi += dt*vp;
	vo += dt*( -A0*(vi+t1*vp) - vo)/t2;

	/* output clipping */
	if (vo >3.0) vo=3.0;
	if (vo <0.2) vo=0.2;
	
	/* vpi_printf("state: %7d %7d %7.1f %g %g %g\n",
		last_time, time->low, dt, vp, vi, vo); */
	last_time=time->low;
}

static void do_filter(struct t_vpi_time *time, int value)
{
	update_state(time);
	vp=value?1.0:-1.0;
}

static void record_state(void) {
	static FILE *analog;
	static int tried_once=0;
	if (!tried_once) {
		analog = fopen("analog.dat","w");
		tried_once=1;
	}
	if (analog) fprintf(analog, "%7d %g %g %g\n",
		last_time, vp, vi, vo);
}

/* picoseconds for a VCO tick */
/* supposed to be centered around 14765 */
static int do_period(struct t_vpi_time *time)
{
	static double residue=0.0;
	double period;
	int iperiod;
	update_state(time);
	period  = 1.e6/emulate_max2605(vo)+residue;
        iperiod = period;          /* truncate, but */
	residue = period-iperiod;  /* remember the truncation for next time */
	record_state();
	return iperiod;
}

/* fluff */
static int period_calltf(char *name)
{
	s_vpi_value val;
	vpiHandle sys, argv, ftime;
	struct t_vpi_time *i_time;

	sys    = vpi_handle(vpiSysTfCall, 0);    assert(sys);
	argv   = vpi_iterate(vpiArgument, sys);  assert(argv);
	ftime  = vpi_scan(argv);                 assert(ftime);

	val.format=vpiTimeVal;
	vpi_get_value(ftime, &val);
	i_time = val.value.time;

	val.format = vpiIntVal;
	val.value.integer = do_period(i_time);
	vpi_put_value(sys, &val, 0, vpiNoDelay);

	return 0;
}

static int filter_calltf(char *name)
{
	s_vpi_value val;
	vpiHandle sys, argv, ftime, fvalue;
	int i_value;
	struct t_vpi_time *i_time;

	sys    = vpi_handle(vpiSysTfCall, 0);    assert(sys);
	argv   = vpi_iterate(vpiArgument, sys);  assert(argv);
	ftime  = vpi_scan(argv);                 assert(ftime);
	fvalue = vpi_scan(argv);                 assert(fvalue);

	val.format=vpiTimeVal;
	vpi_get_value(ftime, &val);
	i_time = val.value.time;

	val.format=vpiIntVal;
	vpi_get_value(fvalue, &val);
	i_value = val.value.integer;

	vpi_free_object(argv);

	do_filter(i_time, i_value);	
	return 0;
}

static void filter_register()
{
	s_vpi_systf_data tf_data;

	tf_data.type      = vpiSysTask;
	tf_data.tfname    = "$filter";
	tf_data.calltf    = filter_calltf;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = 0;
	vpi_register_systf(&tf_data);

	tf_data.type      = vpiSysFunc;
	/* $period is intrinsic in later versions of Icarus */
	tf_data.tfname    = "$vco_period";
	tf_data.calltf    = period_calltf;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = 0;
	vpi_register_systf(&tf_data);
}

void (*vlog_startup_routines[])() = {
	filter_register,
	0
};
